use std::str::FromStr;

use baseline::M3BaselineColors;
use palette::{IntoColor, Lch, Srgb};
use serde::Serialize;
use tonal_palette::TonalPalette;

use crate::convert::map_lch_to_srgb_hex;
use crate::errors;

use super::SchemeExport;

mod baseline;
mod color_set;
mod surface;
mod tonal_palette;

#[derive(Debug, Clone, Serialize)]
pub struct MaterialDesign3Scheme {
    pub white: String,
    pub black: String,
    pub light_baseline: M3BaselineColors,
    pub dark_baseline: M3BaselineColors,
}

impl MaterialDesign3Scheme {
    pub fn new(source_color: &str, error_color: &str) -> Result<Self, errors::ColorError> {
        let source: Lch = Srgb::from_str(source_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(source_color.to_string()))?
            .into_format::<f32>()
            .into_color();
        let error: Lch = Srgb::from_str(error_color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(error_color.to_string()))?
            .into_format::<f32>()
            .into_color();
        let source_hue = source.hue.into_positive_degrees();
        let p = TonalPalette::from_hue_and_chroma(source_hue, source.chroma);
        let s = TonalPalette::from_hue_and_chroma(source_hue, source.chroma / 3.0);
        let t = TonalPalette::from_hue_and_chroma(source_hue + 60.0, source.chroma / 2.0);
        let n = TonalPalette::from_hue_and_chroma(source_hue, (source.chroma / 12.0).min(4.0));
        let nv = TonalPalette::from_hue_and_chroma(source_hue, (source.chroma / 6.0).min(8.0));
        let e = TonalPalette::from_hue_and_chroma(error.hue.into_positive_degrees(), 84.0);

        Ok(Self {
            white: map_lch_to_srgb_hex(&Lch::new(100.0, 0.0, 0.0)),
            black: map_lch_to_srgb_hex(&Lch::new(0.0, 0.0, 0.0)),
            light_baseline: M3BaselineColors::new(&p, &s, &t, &n, &nv, &e, false),
            dark_baseline: M3BaselineColors::new(&p, &s, &t, &n, &nv, &e, true),
        })
    }

    pub fn add_custom_color(
        &mut self,
        name: String,
        color: String,
    ) -> Result<(), errors::ColorError> {
        let custom_color: Lch = Srgb::from_str(&color)
            .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.clone()))?
            .into_format::<f32>()
            .into_color();
        let hue = custom_color.hue.into_positive_degrees();
        let palette = TonalPalette::from_hue_and_chroma(hue, custom_color.chroma);
        self.light_baseline.add_custom_set(name.clone(), &palette);
        self.dark_baseline.add_custom_set(name.clone(), &palette);
        Ok(())
    }
}

impl SchemeExport for MaterialDesign3Scheme {
    fn output_css_variables(&self) -> String {
        let mut css_variables = Vec::new();

        css_variables.push(format!("--color-white: #{};", self.white));
        css_variables.push(format!("--color-black: #{};", self.black));
        css_variables.extend(self.light_baseline.to_css_variables());
        css_variables.extend(self.dark_baseline.to_css_variables());

        css_variables.join("\n")
    }

    fn output_scss_variables(&self) -> String {
        let mut scss_variables = Vec::new();

        scss_variables.push(format!("$color-white: #{};", self.white));
        scss_variables.push(format!("$color-black: #{};", self.black));
        scss_variables.extend(self.light_baseline.to_scss_variables());
        scss_variables.extend(self.dark_baseline.to_scss_variables());

        scss_variables.join("\n")
    }

    fn output_javascript_object(&self) -> String {
        let mut js_object = Vec::new();

        js_object.push("const colorScheme = {".to_string());
        js_object.push(format!("  white: '#{}',", self.white));
        js_object.push(format!("  black: '#{}',", self.black));
        js_object.push("  light: {".to_string());
        js_object.extend(
            self.light_baseline
                .to_javascript_object_fields()
                .into_iter()
                .map(|s| format!("    {}", s))
                .collect::<Vec<String>>(),
        );
        js_object.push("  },".to_string());
        js_object.push("  dark: {".to_string());
        js_object.extend(
            self.dark_baseline
                .to_javascript_object_fields()
                .into_iter()
                .map(|s| format!("    {}", s))
                .collect::<Vec<String>>(),
        );
        js_object.push("  },".to_string());
        js_object.push("}".to_string());

        js_object.join("\n")
    }
}
